/********************************************************************

	Main.c

********************************************************************/

#ifndef JOYSTICKMAIN_C
#define JOYSTICKMAIN_C

/** INCLUDES *******************************************************/
#include "usb.h"
#include "HardwareProfile.h"
#include "joystick1_hid.h"
#include "joystick2_hid.h"

/** CONFIGURATION **************************************************/
// PIC 18F2550 fuse configuration:
// Config word 1 (Oscillator configuration)
// 20Mhz crystal input scaled to 48Mhz and configured for USB operation
__CONFIG(1, USBPLL & IESODIS & FCMDIS & HSPLL & CPUDIV1 & PLLDIV5);
// Config word 2
__CONFIG(2, VREGEN & PWRTDIS & BORDIS & BORV20 & WDTDIS & WDTPS32K);
// Config word 3
__CONFIG(3, PBDIGITAL & LPT1DIS & MCLREN);
// Config word 4
__CONFIG(4, XINSTDIS & STVREN & LVPDIS & ICPORTDIS & DEBUGDIS);
// Config word 5, 6 and 7 (protection configuration)
__CONFIG(5, UNPROTECT);
__CONFIG(6, UNPROTECT);
__CONFIG(7, UNPROTECT);

/** PRIVATE PROTOTYPES *********************************************/
void BlinkUSBStatus(void);
static void InitializeSystem(void);
void ProcessIO(void);
void UserInit(void);
void interrupt YourHighPriorityISRCode();
void interrupt low_priority YourLowPriorityISRCode();
void processJoystick_j1(void);
void processJoystick_j2(void);
void readJoystickState(void);

/** DECLARATIONS ***************************************************/
//http://www.microsoft.com/whdc/archive/hidgame.mspx
#define HAT_SWITCH_NORTH            0x0
#define HAT_SWITCH_NORTH_EAST       0x1
#define HAT_SWITCH_EAST             0x2
#define HAT_SWITCH_SOUTH_EAST       0x3
#define HAT_SWITCH_SOUTH            0x4
#define HAT_SWITCH_SOUTH_WEST       0x5
#define HAT_SWITCH_WEST             0x6
#define HAT_SWITCH_NORTH_WEST       0x7
#define HAT_SWITCH_NULL             0x8

// 35 bytes
typedef union _INTPUT_CONTROLS_TYPEDEF
{
    struct
    {
        struct
        {
            BYTE square:1;
            BYTE x:1;
            BYTE o:1;
            BYTE triangle:1;
            BYTE L1:1;
            BYTE R1:1;
            BYTE L2:1;
            BYTE R2:1;//
            BYTE select:1;
            BYTE start:1;
            BYTE left_stick:1;
            BYTE right_stick:1;
            BYTE home:1;
            BYTE :3;    //filler
        } buttons;
        struct
        {
            BYTE hat_switch:4;
            BYTE :4;//filler
        } hat_switch;
        struct
        {
            BYTE X;
            BYTE Y;
            BYTE Z;
            BYTE Rz;
        } analog_stick;
    } members;
    BYTE val[7];
} INPUT_CONTROLS;

// Structure for holding the joystick switch state
struct switchState {
	unsigned char state; // On or off
	unsigned char debounceCount;
};

// Structure for holding the joystick states
struct joystickState {
	struct switchState joystickSwitch[5];
} joystick[2];

/** VARIABLES ******************************************************/
//char buffer[8];
USB_HANDLE lastTransmission_j1;
USB_HANDLE lastTransmission_j2;

// These variables need to be in the USB RAM
INPUT_CONTROLS joystick_input_j1 @ 0x613; // 35 bytes
INPUT_CONTROLS joystick_input_j2 @ 0x656; // 35 bytes (0x636)
//BYTE hid_report[8] @ 0x659;
	
// Interrupt handling routines.
void interrupt YourHighPriorityISRCode()
{
	//Check which interrupt flag caused the interrupt.
	//Service the interrupt
	//Clear the interrupt flag
	//Etc.
	#if defined(USB_INTERRUPT)
		USBDeviceTasks();
	#endif	
}

void interrupt low_priority YourLowPriorityISRCode()
{
	//Check which interrupt flag caused the interrupt.
	//Service the interrupt
	//Clear the interrupt flag
	//Etc.	
}

// Main function
void main(void)
{
    InitializeSystem();

    #if defined(USB_INTERRUPT)
        USBDeviceAttach();
    #endif

    while(1)
    {
        #if defined(USB_POLLING)
			// Check bus status and service USB interrupts.
	        USBDeviceTasks(); 
        #endif
    				  
		// Application-specific tasks.
        ProcessIO();        
    }
}

// Initialise system
static void InitializeSystem(void)
{
    ADCON1 = 0x0F;                 // Default all pins to digital

	// Configure ports as inputs (1) or outputs(0)
	TRISA = 0b00000000;
	TRISB = 0b00011111;
	TRISC = 0b00000000;

	// Clear all ports
	PORTA = 0b00000000;
	PORTB = 0b00000000;
	PORTC = 0b00000000;

 	// Enable portB weak pull-ups
	RBPU = 0;

    #if defined(USE_USB_BUS_SENSE_IO)
    tris_usb_bus_sense = INPUT_PIN; // See HardwareProfile.h
    #endif
    
    #if defined(USE_SELF_POWER_SENSE_IO)
    tris_self_power = INPUT_PIN;	// See HardwareProfile.h
    #endif
    
    UserInit();

    USBDeviceInit();	//usb_device.c.  Initializes USB module SFRs and firmware
    					//variables to known states.
}

// User initialisation
void UserInit(void)
{
	int loop;

    // Initialise the status LED
	STATUS_LED = 0;
    
	// Initialise the joystick states
	for (loop = 0; loop < 5; loop++)
	{
		joystick[PORT1].joystickSwitch[loop].state = OFF;
		joystick[PORT1].joystickSwitch[loop].debounceCount = 0;

		joystick[PORT2].joystickSwitch[loop].state = OFF;
		joystick[PORT2].joystickSwitch[loop].debounceCount = 0;
	}

    // Initialize the variables holding the handles for the last transmission
    lastTransmission_j1 = 0;
    lastTransmission_j2 = 0;
}


// Process input and output
void ProcessIO(void)
{   
    //Blink the LEDs according to the USB device status
    BlinkUSBStatus();

    // User Application USB tasks
    if((USBDeviceState < CONFIGURED_STATE)||(USBSuspendControl==1)) return;

	// Read and debounce the joystick switches
	readJoystickState();

	// Call the joystick USB communication functions
	processJoystick_j1();
	processJoystick_j2();
}

// Read the joystick switch states and perform debounce
void readJoystickState(void)
{
	unsigned char port;
	
	for (port = 0; port < 2; port++)
	{
		// Select the correct joystick for the next pass...

		// The physical joystick ports are multiplexed and are selected
		// by sinking the current either through RC0 (port 1) or RC1 (port 2)
		// You switch them by changing RC0 and RC1 to either inputs (which
		// are high-impedence and won't sink current) or outputs (which will).
		if (port == 0)
		{
			TRISC0 = 0;
			TRISC1 = 1;
			JOYSTICK1 = 0;
			JOYSTICK2 = 0;
			__delay_us(2);

		}
		else
		{
			TRISC0 = 1;
			TRISC1 = 0;
			JOYSTICK1 = 0;
			JOYSTICK2 = 0;
			__delay_us(2);
		}
	
		// Debounce FORWARD
		if (JOYSTICK_FORWARD == 0 && joystick[port].joystickSwitch[FORWARD].state == OFF)
			if (joystick[port].joystickSwitch[FORWARD].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[FORWARD].state = ON;
		else joystick[port].joystickSwitch[FORWARD].debounceCount++;
		
		if (JOYSTICK_FORWARD == 1 && joystick[port].joystickSwitch[FORWARD].state == ON)
			if (joystick[port].joystickSwitch[FORWARD].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[FORWARD].state = OFF;
		else joystick[port].joystickSwitch[FORWARD].debounceCount++;
		
		if (JOYSTICK_FORWARD != joystick[port].joystickSwitch[FORWARD].state)
			joystick[port].joystickSwitch[FORWARD].debounceCount = 0;
		
		// Debounce BACKWARD
		if (JOYSTICK_BACKWARD == 0 && joystick[port].joystickSwitch[BACKWARD].state == OFF)
			if (joystick[port].joystickSwitch[BACKWARD].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[BACKWARD].state = ON;
		else joystick[port].joystickSwitch[BACKWARD].debounceCount++;
		
		if (JOYSTICK_BACKWARD == 1 && joystick[port].joystickSwitch[BACKWARD].state == ON)
			if (joystick[port].joystickSwitch[BACKWARD].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[BACKWARD].state = OFF;
		else joystick[port].joystickSwitch[BACKWARD].debounceCount++;
		
		if (JOYSTICK_BACKWARD != joystick[port].joystickSwitch[BACKWARD].state)
			joystick[port].joystickSwitch[BACKWARD].debounceCount = 0;
		
		// Debounce LEFT
		if (JOYSTICK_LEFT == 0 && joystick[port].joystickSwitch[LEFT].state == OFF)
			if (joystick[port].joystickSwitch[LEFT].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[LEFT].state = ON;
		else joystick[port].joystickSwitch[LEFT].debounceCount++;
		
		if (JOYSTICK_LEFT == 1 && joystick[port].joystickSwitch[LEFT].state == ON)
			if (joystick[port].joystickSwitch[LEFT].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[LEFT].state = OFF;
		else joystick[port].joystickSwitch[LEFT].debounceCount++;
		
		if (JOYSTICK_LEFT != joystick[port].joystickSwitch[LEFT].state)
			joystick[port].joystickSwitch[LEFT].debounceCount = 0;
		
		// Debounce RIGHT
		if (JOYSTICK_RIGHT == 0 && joystick[port].joystickSwitch[RIGHT].state == OFF)
			if (joystick[port].joystickSwitch[RIGHT].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[RIGHT].state = ON;
		else joystick[port].joystickSwitch[RIGHT].debounceCount++;
		
		if (JOYSTICK_RIGHT == 1 && joystick[port].joystickSwitch[RIGHT].state == ON)
			if (joystick[port].joystickSwitch[RIGHT].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[RIGHT].state = OFF;
		else joystick[port].joystickSwitch[RIGHT].debounceCount++;
		
		if (JOYSTICK_RIGHT != joystick[port].joystickSwitch[RIGHT].state)
			joystick[port].joystickSwitch[RIGHT].debounceCount = 0;
		
		// Debounce FIRE
		if (JOYSTICK_FIRE == 0 && joystick[port].joystickSwitch[FIRE].state == OFF)
			if (joystick[port].joystickSwitch[FIRE].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[FIRE].state = ON;
		else joystick[port].joystickSwitch[FIRE].debounceCount++;
		
		if (JOYSTICK_FIRE == 1 && joystick[port].joystickSwitch[FIRE].state == ON)
			if (joystick[port].joystickSwitch[FIRE].debounceCount > JOYSTICK_DEBOUNCE)
				joystick[port].joystickSwitch[FIRE].state = OFF;
		else joystick[port].joystickSwitch[FIRE].debounceCount++;
		
		if (JOYSTICK_FIRE != joystick[port].joystickSwitch[FIRE].state)
			joystick[port].joystickSwitch[FIRE].debounceCount = 0;
	}
}

// Process joystick 1
void processJoystick_j1(void)
{
    // If the last transmision is complete send the joystick status
    if(!HIDTxHandleBusy(lastTransmission_j1))
    {
		// Set everything to the defult
		joystick_input_j1.members.buttons.x = 0;
		joystick_input_j1.members.buttons.square = 0;
		joystick_input_j1.members.buttons.o = 0;
		joystick_input_j1.members.buttons.triangle = 0;
		joystick_input_j1.members.buttons.L1 = 0;
		joystick_input_j1.members.buttons.R1 = 0;
		joystick_input_j1.members.buttons.L2 = 0;
		joystick_input_j1.members.buttons.R2 = 0;
		joystick_input_j1.members.buttons.select = 0;
		joystick_input_j1.members.buttons.start = 0;
		joystick_input_j1.members.buttons.left_stick = 0;
		joystick_input_j1.members.buttons.right_stick = 0;
		joystick_input_j1.members.buttons.home = 0;
		joystick_input_j1.members.hat_switch.hat_switch = HAT_SWITCH_NULL;

		// 128 is 'middle'
		joystick_input_j1.members.analog_stick.X = 128;
		joystick_input_j1.members.analog_stick.Y = 128;

		// Now adjust according to the joystick switch states
		if (joystick[PORT1].joystickSwitch[FORWARD].state == ON)
			joystick_input_j1.members.analog_stick.Y = 0;

		if (joystick[PORT1].joystickSwitch[BACKWARD].state == ON)
			joystick_input_j1.members.analog_stick.Y = 255;

		if (joystick[PORT1].joystickSwitch[RIGHT].state == ON)
			joystick_input_j1.members.analog_stick.X = 255;

		if (joystick[PORT1].joystickSwitch[LEFT].state == ON)
			joystick_input_j1.members.analog_stick.X = 0;

		if (joystick[PORT1].joystickSwitch[FIRE].state == ON)
			joystick_input_j1.members.buttons.square = 1;

		//Send the packet over USB to the host.
		lastTransmission_j1 = HIDTxPacket(HID_EP_J1, (BYTE*)&joystick_input_j1, sizeof(joystick_input_j1));
    }	
}

// Process joystick 2
void processJoystick_j2(void)
{
    // If the last transmision is complete send the joystick status
    if(!HIDTxHandleBusy(lastTransmission_j2))
    {
		// Set everything to the defult
		joystick_input_j2.members.buttons.x = 0;
		joystick_input_j2.members.buttons.square = 0;
		joystick_input_j2.members.buttons.o = 0;
		joystick_input_j2.members.buttons.triangle = 0;
		joystick_input_j2.members.buttons.L1 = 0;
		joystick_input_j2.members.buttons.R1 = 0;
		joystick_input_j2.members.buttons.L2 = 0;
		joystick_input_j2.members.buttons.R2 = 0;
		joystick_input_j2.members.buttons.select = 0;
		joystick_input_j2.members.buttons.start = 0;
		joystick_input_j2.members.buttons.left_stick = 0;
		joystick_input_j2.members.buttons.right_stick = 0;
		joystick_input_j2.members.buttons.home = 0;
		joystick_input_j2.members.hat_switch.hat_switch = HAT_SWITCH_NULL;

		// 128 is 'middle'
		joystick_input_j2.members.analog_stick.X = 128;
		joystick_input_j2.members.analog_stick.Y = 128;

		// Now adjust according to the joystick switch states
		if (joystick[PORT2].joystickSwitch[FORWARD].state == ON)
			joystick_input_j2.members.analog_stick.Y = 0;

		if (joystick[PORT2].joystickSwitch[BACKWARD].state == ON)
			joystick_input_j2.members.analog_stick.Y = 255;

		if (joystick[PORT2].joystickSwitch[RIGHT].state == ON)
			joystick_input_j2.members.analog_stick.X = 255;

		if (joystick[PORT2].joystickSwitch[LEFT].state == ON)
			joystick_input_j2.members.analog_stick.X = 0;

		if (joystick[PORT2].joystickSwitch[FIRE].state == ON)
			joystick_input_j2.members.buttons.square = 1;

		//Send the packet over USB to the host.
		lastTransmission_j2 = HIDTxPacket(HID_EP_J2, (BYTE*)&joystick_input_j2, sizeof(joystick_input_j2));
    }	
}

// Blink the USB status LED
void BlinkUSBStatus(void)
{
    static WORD led_count=0;
    
    if(led_count == 0)led_count = 10000U;
    led_count--;

    if(USBDeviceState == CONFIGURED_STATE)
    {
		if(led_count==0)
		{
			// Uncomment this if you want the status light to blink
			// Toggle the status LED
			// STATUS_LED = !STATUS_LED;

			// Or use this for a (less annoying steady light)
			STATUS_LED = 1;
		}
    }
}

// Call back that is invoked when a USB suspend is detected
void USBCBSuspend(void)
{
}

// This function is called when the USB interrupt bit is set
// In this example the interrupt is only used when the device
// goes to sleep when it receives a USB suspend command
#if 0
void __attribute__ ((interrupt)) _USB1Interrupt(void)
{
    #if !defined(self_powered)
        if(U1OTGIRbits.ACTVIF)
        {       
            IEC5bits.USB1IE = 0;
            U1OTGIEbits.ACTVIE = 0;
            IFS5bits.USB1IF = 0;
        
            //USBClearInterruptFlag(USBActivityIFReg,USBActivityIFBitNum);
            USBClearInterruptFlag(USBIdleIFReg,USBIdleIFBitNum);
            //USBSuspendControl = 0;
        }
    #endif
}
#endif
	
// This call back is invoked when a wakeup from USB suspend is detected.
void USBCBWakeFromSuspend(void)
{
}

// The USB host sends out a SOF packet to full-speed devices every 1 ms.
void USBCB_SOF_Handler(void)
{
    // No need to clear UIRbits.SOFIF to 0 here.
    // Callback caller is already doing that.
}

// The purpose of this callback is mainly for
// debugging during development. Check UEIR to see
// which error causes the interrupt.
void USBCBErrorHandler(void)
{
    // No need to clear UEIR to 0 here.
    // Callback caller is already doing that.
}

// Check other (non-standard) requests
void USBCBCheckOtherReq(void)
{
    // Send request to the interface specified by wIndex
    switch ( SetupPkt.wIndex ) {
        case HID_INTF_ID_J1: USBCheckHIDRequest_j1();    break;
        case HID_INTF_ID_J2: USBCheckHIDRequest_j2();    break;
        default:                                         break;
    }
}

// SET_DESCRIPTOR request
void USBCBStdSetDscHandler(void)
{
    // Must claim session ownership if supporting this request
}

// Initialise end-points
void USBCBInitEP(void)
{
    // Enable the Joystick HID endpoints for J1 and J2
    USBEnableEndpoint(HID_EP_J1,USB_OUT_ENABLED|USB_IN_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);
    USBEnableEndpoint(HID_EP_J2,USB_OUT_ENABLED|USB_IN_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);
}

// Send resume
void USBCBSendResume(void)
{
    static WORD delay_count;
    
    USBResumeControl = 1;                // Start RESUME signaling
    
    delay_count = 1800U;                // Set RESUME line for 1-13 ms
    do
    {
        delay_count--;
    }while(delay_count);
    USBResumeControl = 0;
}

// USB Callback event handler
BOOL USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, WORD size)
{
    switch(event)
    {
        case EVENT_CONFIGURED: 
            USBCBInitEP();
            break;
        case EVENT_SET_DESCRIPTOR:
            USBCBStdSetDscHandler();
            break;
        case EVENT_EP0_REQUEST:
            USBCBCheckOtherReq();
            break;
        case EVENT_SOF:
            USBCB_SOF_Handler();
            break;
        case EVENT_SUSPEND:
            USBCBSuspend();
            break;
        case EVENT_RESUME:
            USBCBWakeFromSuspend();
            break;
        case EVENT_BUS_ERROR:
            USBCBErrorHandler();
            break;
        case EVENT_TRANSFER:
            Nop();
            break;
        default:
            break;
    }      
    return TRUE; 
}

#endif
